div#python2-eol {
	border-color: red;
	border-width: medium;
}

/* Ensure minimum width for 'Parameters' / 'Returns' column */
dl.field-list > dt {
    min-width: 100px
}

/* Insert space between methods for readability */
dl.method {
	padding-top: 10px;
	padding-bottom: 10px
}

/* Insert empty space between classes */
dl.class {
	padding-bottom: 50px
}
